/*
 *   This file is part of the CxProlog system

 *   Dictionary.c
 *   by A.Miguel Dias - 2000/09/29
 *   CITI - Centro de Informatica e Tecnologias da Informacao
 *   Dept. de Informatica, FCT, Universidade Nova de Lisboa.
 *   Copyright (C) 1990-2001 A.Miguel Dias, CITI, DI/FCT/UNL

 *   CxProlog is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.

 *   CxProlog is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *   GNU General Public License for more details.

 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "CxProlog.h"

typedef struct Dict
{
	Word tagHolder ;
	Hdl begin, end, first, last  ;
	struct Dict *next ;
} Dict, *DictPt ;

#define DictTagHolder(d)		((d)->tagHolder)
#define DictBegin(d)			((d)->begin)
#define DictEnd(d)				((d)->end)
#define DictFirst(d)			((d)->first)
#define DictLast(d)				((d)->last)
#define DictNext(d)				((d)->next)

#define cDictPt(x)				((DictPt)(x))

static DictPt dictList = nil ;
static DictPt freeDictList = nil ;

static void DictInit(DictPt d, Size capacity)
{
	DictBegin(d) = TempBlockAllocate(capacity * 2) ;
	DictEnd(d) = DictBegin(d) + capacity * 2 ;
	DictFirst(d) = DictBegin(d) ;
	DictLast(d) = DictBegin(d) ;
}

static Size DictSize(DictPt d)
{
	return (DictLast(d) - DictFirst(d)) / 2 ;
}

static Size DictCapacity(DictPt d)
{
	return (DictEnd(d) - DictBegin(d)) / 2 ;
}

static void DictExpand(DictPt d)
{
	Hdl b = DictBegin(d) ;
	Hdl f = DictFirst(d) ;
	Hdl l = DictLast(d) ;
	Hdl h ;
	
	DictInit(d, 2 * DictCapacity(d)) ;
	for( h = DictBegin(d) ; f < l ; *h++ = *f++ ) ;
	DictLast(d) = h ;
	BlockRelease(b) ;
}

static DictPt DictNew(void)
{
	DictPt d = TempBlockAllocate(WordsOf(Dict)) ;
	DictTagHolder(d) = dictSubTag ;
	DictNext(d) = dictList ;
	dictList = d ;
	DictInit(d, 4) ;
	return d ;
}

static void DictClear(DictPt d)
{
	register Hdl h ;
	for( h = DictFirst(d) ; h < DictLast(d) ; h += 2 ) {
		ReleaseTerm(h[0]) ;
		ReleaseTerm(h[1]) ;
	}
	DictLast(d) = DictFirst(d) ;
}

static void DictRelease(DictPt d)
{
	DictClear(d) ;
	BlockRelease(DictBegin(d)) ;
	DictBegin(d) = nil ;	/* Mark "dictionary not in use" */
	DictNext(d) = freeDictList ;
	freeDictList = d ;
}

static void DictDelete(DictPt d)
{
	if( dictList == d )	/* first is removed */
		dictList = DictNext(d) ;
	else {
		register DictPt x ;
		for( x = dictList ; DictNext(x) != d ; x = DictNext(x) ) ;
		DictNext(x) = DictNext(d) ;
	}
	DictRelease(d) ;
}

static void DictSet(DictPt d, Pt k, Pt t)
{
	register Hdl h ;
	k = Drf(k) ;
	for( h = DictFirst(d) ; h < DictLast(d) ; h += 2 )
		if( Compare(*h, k) == 0 ) {
			ReleaseTerm(h[1]) ;
			h[1] = AllocateTermForAssign(t) ;
			return ;
		}
	if( DictLast(d) == DictEnd(d) )
		DictExpand(d) ;
	*DictLast(d)++ = AllocateTermForAssign(k) ;
	*DictLast(d)++ = AllocateTermForAssign(t) ;
}

static Bool DictGet(DictPt d, Pt k, Pt *t)
{
	register Hdl h ;
	k = Drf(k) ;
	for( h = DictFirst(d) ; h < DictLast(d) ; h += 2 )
		if( Compare(*h, k) == 0 ) {
			*t = h[1] ;
			return true ;
		}
	return false ;
}

static Bool DictDeleteItem(DictPt d, Pt k)
{
	register Hdl h ;
	k = Drf(k) ;
	for( h = DictFirst(d) ; h < DictLast(d) ; h += 2 )
		if( Compare(*h, k) == 0 ) {
			ReleaseTerm(h[0]) ;
			ReleaseTerm(h[1]) ;
			for( ; h < DictLast(d) - 2 ; h += 2 ) {
				h[0] = h[2] ;
				h[1] = h[3] ;
			}
			DictLast(d) -= 2 ;
			return true ;
		}
	return false ;
}

static void DictWrite(StreamPt stm, DictPt d)
{
	register Hdl h ;
	StreamWrite(stm, "%s", XExtraAsStr(TagExtra(d))) ;
	StreamWrite(stm, "     (current capacity %ld)\n", DictCapacity(d)) ;
	for( h = DictFirst(d) ; h < DictLast(d) ; h += 2 ) {
		StreamWrite(stm, "\t%s -- ", TermAsStr(h[0])) ;
		StreamWrite(stm, "%s\n", TermAsStr(h[1])) ;
	}
}

static DictPt UsingDict(DictPt d)
{
	if( DictBegin(d) == nil )
		ImperativeError("Invalid operation over deleted dictionary %s",
									XExtraAsStr(TagExtra(d))) ;
	return d ;
}

Bool DictCheck(VoidPt ref)
{
	register DictPt d ;
	dolist(d, dictList, DictNext(d))
		if( d == ref )
			return DictBegin(d) != nil ;
	return false ;
}

void DictsAtomGCMark()
{
	register DictPt d ;
	register Hdl h ;
	dolist(d, dictList, DictNext(d))
		if( DictBegin(d) != nil )
			for( h = DictFirst(d) ; h < DictLast(d) ; h += 2 ) {
				TermAtomGCMark(h[0]) ;
				TermAtomGCMark(h[1]) ;
			}
}


/* CXPROLOG C'BUILTINS */

static DictPt XTestDict(register Pt t)
{
	VarValue(t) ;
	if( IsAtom(t) )
		t = IVarGet(XAtom(t)) ;
	if( IsExtra(t) && XExtraSubTag(t) == dictSubTag )
		return UsingDict(cDictPt(XPt(t))) ;
	TypeError2("DICT or IVAR", t) ;
	return nil ;
}

static void PDictCheck()
{
	Pt t = Drf(X0) ;
	if( IsAtom(t) )
		t = IVarGet(XAtom(t)) ;
	if( IsExtra(t) && XExtraSubTag(t) == dictSubTag && DictCheck(XPt(t)) )
		JumpNext()
	DoFail()
}

static void PDictNew()
{
	Pt t = Drf(X0) ;
	if( IsVar(t) ) {
		DictPt d = DictNew() ;
		if( UnifyWithAtomic(t, TagExtra(d)) ) JumpNext()
		InternalError("PDictNew") ;
	}
	if( IsAtom(t) ) {
		DictPt d = DictNew() ;
		IVarSet(XAtom(t), TagExtra(d)) ;
		JumpNext()
	}
	TypeError2("VAR or IVAR", t) ;
}

static void PDictClear()
{
	DictClear(XTestDict(X0)) ;
	JumpNext()
}

static void PDictDelete()
{
	DictDelete(XTestDict(X0)) ;
	JumpNext()
}

static void PDictSet()
{
	DictSet(XTestDict(X0),XTestNonVar(X1),X2) ;
	JumpNext()
}

static void PDictGet()
{
	Pt t ;
	if( DictGet(XTestDict(X0),XTestNonVar(X1),&t) ) {
		t = ZPushTerm(t) ;
		if( Unify(X2,t) ) JumpNext()
	}
	DoFail()
}

static void PDictDeleteItem()
{
	if( DictDeleteItem(XTestDict(X0),XTestNonVar(X1)) ) JumpNext()
	DoFail()
}

static void PDictAsList()
{
	DictPt d = XTestDict(X0) ;
	Hdl h ; Pt t ;
	Word w[3] ;
	w[0] = cWord(hifenFunctor) ;	
	Z = tNilAtom ;
	for( h = DictLast(d) - 2 ; h >= DictFirst(d) ; h -= 2 ) {
		w[1] = cWord(h[0]) ;
		w[2] = cWord(h[1]) ;
		t = ZPushTerm(TagStruct(w)) ;
		Z = MakeBinStruct(listFunctor, t, Z) ;
	}
	if( Unify(Z, X1) ) JumpNext()
	DoFail()
}

static void PDictWrite()
{
	DictWrite(currOut, XTestDict(X0)) ;
	JumpNext()
}

static void PSDictWrite()
{
	DictWrite(XTestStream(X0, mWrite), XTestDict(X1)) ;
	JumpNext()
}

static void PNDCurrentDict()
{
	DictPt d =
		A(1) == tNilAtom ? dictList : DictNext(cDictPt(A(1))) ;
	A(1) = cPt(d) ;
	if( d == nil ) Jump(DiscardAndFail) ;
	if( Unify(X0, TagExtra(d)) ) JumpNext()
	DoFail()
}

static void PDicts()
{
	DictPt d ;
	VersionShow() ;
	Write("Dicts:\n") ;
	dolist(d, dictList, DictNext(d)) {
		AtomPt at = IVarWith(TagExtra(d)) ;
		Write("  %16s -> size =%2d, capacity =%2d", 
					TermAsStr(TagExtra(d)),
					DictSize(d),
					DictCapacity(d)) ;
		if( at != nil )
			Write(" (in ivar '%s')", AtomName(at)) ;
		Write("\n") ;
	}
	JumpNext()
}

void DictsInit()
{
	IVarConstSet(LookupAtom("$dict"), TagExtra(DictNew())) ;

	InstallCBuiltinPred("dict", 1, PDictCheck) ;
	InstallCBuiltinPred("dict_new", 1, PDictNew) ;
	InstallCBuiltinPred("dict_clear", 1, PDictClear) ;
	InstallCBuiltinPred("dict_delete", 1, PDictDelete) ;
	InstallCBuiltinPred("dict_set", 3, PDictSet) ;
	InstallCBuiltinPred("dict_get", 3, PDictGet) ;
	InstallCBuiltinPred("dict_delete_item", 2, PDictDeleteItem) ;
	InstallCBuiltinPred("dict_as_list", 2, PDictAsList) ;
	InstallCBuiltinPred("dict_write", 1, PDictWrite) ;
	InstallCBuiltinPred("dict_write", 2, PSDictWrite) ;
	InstallNDeterCBuiltinPred("current_dict", 1, PNDCurrentDict) ;
	InstallNDeterCBuiltinPred("dicts", 0, PDicts) ;
}
